// IFF System

//Responder Module:
//    setup_send: Checks the frequency of interrupts from the input sensors, and sends a "friendly-fire" signal to the initiator unit. Once the frequency falls outside the valid range, it sends another "cease" siganl to the initiator unit to indicate that it can resume normal operation
//    plethys: Sets up the analog interrupts for the plethysmograph, and checks if there is a pulse rate within the last 4 seconds. If not, light up an LED to indicate system will be unresponsive (in the field, the system will erase its memory)
//    check_plethys: A task to debounce the analog interrupts from the plethysmograph

//Initiator Module:
//    laser_recv: sets up the laser signal and checks periodically for a friendly-fire signal from the responder unit.

#include "trtSettings.h"
#include "trtkernel_1284.c"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// serial communication library
// Don't mess with the semaphores
#define SEM_RX_ISR_SIGNAL 1
#define SEM_STRING_DONE 2 // user hit <enter>
#include "trtUart.h"
#include "trtUart.c"

#define LASER	//comment out to program for receive module

const int8_t LCD_initialize[] PROGMEM = "LCD Initialized\0";
const int8_t LCD_line[] PROGMEM = "line 1\0";
const int8_t LCD_number[] PROGMEM = "Number=\0";
int8_t lcd_status[17];	// LCD display buffer
int8_t lcd_period[17];	

volatile uint16_t laser_period, tmp_period;
volatile uint16_t laser_period_ovlf;


// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

// semaphore to protect shared variable
#define SEM_SHARED 3

// the usual
#define begin {
#define end }

#define SET(U, N) ((void)((U) |= 1u << (N)))
#define CLR(U, N) ((void)((U) &= ~(1u << (N))))


char count = 0x00;

unsigned char led;	 //light states 
unsigned char pulse_count;
unsigned char pulse_rate;


// input arguments to each thread
// not actually used in this example
int args[4] ;

// --- debouncing for plethysmograph. re-enable analog compare 10ms after the first trigger-----------------------------
void check_plethys(void* args){
    uint32_t rel, dead ;
    while (1) {

        if (pulse_count > 0) {
            _delay_ms(10);
            ACSR |= (1<<ACIE) | (0<<ACBG) | (1<<ACIS1) | (1<<ACIS0 );	//Enable analog comparator interrupt
        }
        rel = trtCurrentTime() + SECONDS2TICKS(0.1);
        dead = trtCurrentTime() + SECONDS2TICKS(0.1);
        trtSleepUntil(rel, dead);
    }

}

// --- Checks the frequency of interrupts from the input sensors, and sends a "friendly-fire" signal to the initiator unit. Once the frequency falls outside the valid range, it sends another "cease" siganl to the initiator unit to indicate that it can resume normal operation  ----------------------------------------
void setup_send(void* args) 
{	

    uint32_t rel, dead ;
    uint8_t valid, prev_valid;
    
    prev_valid = 0;

    while(1)
    {
        valid = laser_period < 2600 && laser_period > 2400;
        
        if (valid != prev_valid) {
            trtWait(SEM_SHARED) ;
            if (valid == 1) {
                fprintf(stdout, "s1\n\r") ;
            } else {
                fprintf(stdout, "s0\n\r") ;
            }
            trtSignal(SEM_SHARED);
            prev_valid = valid;
        }
        rel = trtCurrentTime() + SECONDS2TICKS(0.1);
        dead = trtCurrentTime() + SECONDS2TICKS(0.1);
        trtSleepUntil(rel, dead);
    }

}

// --- sets up the laser signal and checks periodically for a friendly-fire signal from the responder unit. ----------------------------------------
void laser_recv(void* args) 
{
    uint32_t rel, dead ;
    int input ;
    char cmd[100] ;

    TCCR3B |=  (1<<CS31) | (1<<CS30) | (1<<WGM32) | (1<<WGM33);
    TCCR3A |= (1<<COM0A1) | (1<<WGM31);

    ICR3 = 5000;	//period
    OCR3A = 2500;

    //turn on pwm 
    // turn on fast PWM and OC0A output
    // at full clock rate, toggle OC0A (pin B3) 
    // 16 microsec per PWM cycle sample time

    TCCR0A = (1<<COM0A1) | (1<<WGM00) | (1<<WGM01) ; 	  

    TCCR0B = (1<<CS02) ; //256 prescalar 
    OCR0A = 0x10;

    //turn on 0C3A
    DDRB |= (1<<PINB6);

    while(1)
    {

        fscanf(stdin, "%s", cmd) ;
        trtWait(SEM_STRING_DONE);

        if ( strlen(cmd) > 1 && cmd[0] == 's') {
            
            if (cmd[1] == '1') {
                DDRB |= (1<<PINB3) ;
            } else {
                CLR(DDRB, PINB3);
            }
        } 

        rel = trtCurrentTime() + SECONDS2TICKS(0.03);
        dead = trtCurrentTime() + SECONDS2TICKS(0.03);
        trtSleepUntil(rel, dead);

    }
}



//========================================================================
//Sets up the analog interrupts for the plethysmograph, and checks if there is a pulse rate within the last 4 seconds. If not, light up an LED to indicate system will be unresponsive (in the field, the system will erase its memory)
void plethys(void* args) 
{
    uint32_t rel, dead ;
    uint8_t alive = 1;
    pulse_count = 1;
    
    
    
    PORTC |= (1<<PINC6);
    //set up Analog Comparator
    ACSR |= (1<<ACIE) | (0<<ACBG) | (0<<ACIS1) | (0<<ACIS0 );	//Enable analog comparator interrupt  
    while (1) {

        trtWait(SEM_SHARED) ;
        if ( alive == 1 && pulse_count > 0 )
        {
            fprintf(stdout,"%d\n\r" , pulse_count); 
            
        }
        else
        {
            fprintf(stdout,"DEAD:%d\n\r" , pulse_count); 
            alive = 0;
            DDRC |= (1<<PINC6);
            PORTC |= (1<<PINC6);
            
        }
        trtSignal(SEM_SHARED);
        
        pulse_count = 0 ; 

        rel = trtCurrentTime() + SECONDS2TICKS(4);
        dead = trtCurrentTime() + SECONDS2TICKS(4);
        trtSleepUntil(rel, dead); 
    }
}  


#ifndef LASER
//========================================================================
//Analog Comparator Input ISR
ISR (ANALOG_COMP_vect)
{      
    //update with latest reading
    pulse_count += 1 ;

    ACSR = 0;

}  
#endif


// --- external interrupt ISR ------------------------
ISR (INT0_vect) {
    tmp_period = TCNT2 + laser_period_ovlf  ;      
    if (tmp_period > 10 ) laser_period = tmp_period; // to avoid noise
    TCNT2 = 0 ;
    laser_period_ovlf = 0 ;

}

#ifndef LASER
// --- set up extra 8 bits on timer 2 ----------------
ISR (TIMER2_OVF_vect) {
    laser_period_ovlf = laser_period_ovlf + 256 ;
    if (laser_period_ovlf > 0xDFFF) laser_period = 0xDFFF;
}
#endif

// --- Main Program ----------------------------------
int main(void) {
    
    //init the UART -- trt_uart_init() is in trtUart.c
    trt_uart_init();
    stdout = stdin = stderr = &uart_str;
    fprintf(stdout,"Laser Control\n\r");

    //set up INT0 - pin D2
    EIMSK = 1<<INT0 ; // turn on int0
    EICRA = 3 ;       // rising edge

    DDRD &= 0xFB;
    PORTD &= 0xFB;
    
    #ifndef LASER
    // timer 2 - Counts the period of received signal
    // turn on timer 2 to be read in int0 ISR
    TCCR2B = 5 ; // divide by 128
    //TCCR2B = 4 ; // divide by 64
    //TCCR2B = 7 ; // divide by 1024
    // turn on timer 2 overflow ISR for double precision time
    TIMSK2 = 1 ;

    #endif
    
    // start TRT
    trtInitKernel(80); // 80 bytes for the idle task stack

    // --- create semaphores ----------
    // You must creat the first two semaphores if you use the uart
    trtCreateSemaphore(SEM_RX_ISR_SIGNAL, 0) ; // uart receive ISR semaphore
    trtCreateSemaphore(SEM_STRING_DONE,0) ;  // user typed <enter>

    // variable protection
    trtCreateSemaphore(SEM_SHARED, 1) ; // protect shared variables

    // --- creat tasks  ----------------
    
    #ifndef LASER
    trtCreateTask(setup_send, 200, SECONDS2TICKS(0.05), SECONDS2TICKS(0.05), &(args[0]));
    trtCreateTask(plethys, 200, SECONDS2TICKS(4), SECONDS2TICKS(4), &(args[2]));	
    trtCreateTask(check_plethys, 200, SECONDS2TICKS(4), SECONDS2TICKS(4), &(args[3]));	
    #endif

    #ifdef LASER
    trtCreateTask(laser_recv, 200, SECONDS2TICKS(0.1), SECONDS2TICKS(0.1), &(args[1]));
    #endif
    

    // --- Idle task --------------------------------------
    // just sleeps the cpu to save power 
    // every time it executes
    set_sleep_mode(SLEEP_MODE_IDLE);
    sleep_enable();
    while (1) 
    {
        sleep_cpu();
    }

} // main
